--[[
Name: SpecialEvents-Loot-1.0
Revision: $Rev: 17337 $
Author: Xuerian (sky.shell@gmail.com)
Website: http://www.wowace.com/
Description: Special events for players looting items.
Dependencies: AceLibrary, AceEvent-2.0, Deformat
]]

local vmajor, vminor = "SpecialEvents-Loot-1.0", "$Revision: 17337 $"

if not AceLibrary then error(vmajor .. " requires AceLibrary.") end
if not AceLibrary:HasInstance("AceEvent-2.0") then error(vmajor .. " requires AceEvent-2.0.") end
if not AceLibrary:HasInstance("Deformat-2.0") then error(vmajor .. " requires Deformat-2.0.") end
if not AceLibrary:IsNewVersion(vmajor, vminor) then return end

local lib = {}
AceLibrary("AceEvent-2.0"):embed(lib)
local deformat = AceLibrary("Deformat-2.0")


----------------------------
--     Initializaiton    --
----------------------------

-- Yay debug.
local Dump = DevTools_Dump or function() end

-- Activate a new instance of this library
function activate(self, oldLib, oldDeactivate)
	if oldLib then
		self.vars = oldLib.vars
		oldLib:UnregisterAllEvents()
	else
		self.vars = { }
	end
	
	-- Hook up with chat events for looting
	self:RegisterEvent("CHAT_MSG_LOOT", "Handler")
	self:RegisterEvent("CHAT_MSG_MONEY", "Handler")
	
	-- Register for the loopback, called in XLoot Monitor's options, mainly
	self:RegisterEvent("SpecialEvents_LootDummy", "Handler")
	
	-- Incriment and deincement rolls to only match while there is a roll happening
	self.rolls = 0
	self:RegisterEvent("START_LOOT_ROLL", function() self.rolls = self.rolls + 1 end)
	self:RegisterEvent("CANCEL_LOOT_ROLL", function() self.rolls = self.rolls < 1 and 0 or self.rolls - 1 end)

	-- Hardcoded coin value tables to parse coin strings against
	self.coinage = { { GOLD, 10000 }, { SILVER, 100 }, { COPPER, 1 } }
	
	-- Counter between loot patern sorts
	self.unsortedloot = 0
	
	-- Hardcode loot patterns with starting mode counters
	-- [1] Self count [2] Group count [3] Pattern [4 key] Recipient [5 key] Item [6 key] Count
	self.patterns = {
		{ 0, 0, YOU_LOOT_MONEY, 'self', 'coin', 1 },
		{ 0, 0, LOOT_MONEY_SPLIT, 'self', 'coin', 1 },
		{ 0, 0, LOOT_ITEM_MULTIPLE, 'self', 1, 2, 3 },
		{ 0, 0, LOOT_ITEM, 1, 2 },
		{ 0, 0, LOOT_ITEM_SELF_MULTIPLE, 'self', 1, 2 },
		{ 0, 0, LOOT_ITEM_SELF, 'self', 1 },
		{ 0, 0, LOOT_ITEM_PUSHED_SELF_MULTIPLE, 'self', 1, 2 },
		{ 0, 0, LOOT_ITEM_PUSHED_SELF, 'self', 1 },
	}
	
	-- Generate group loot mode handlers
	self.rollhandlers = { }

	-- Base handler function. Checks argument types, adds to handler table (Presorted)
	--    The pattern is stored as the first key, and the function to be called in the second
	local handler = function(str, func)
		assert(type(str) == "string", "First argument of handler() must be a string, was a "..type(str))
		assert(type(func) == "function" or not func, "Second argument of handler() must be a function, and must exist")
		table.insert(self.rollhandlers, { str, func })
	end
	
	-- Figures the appropriate owner of the event
	local ncheck = function(own, who) 
		if own or  who == YOU then 
			return UnitName('player')
		end
		return who
	end
	
	-- Roll selected handler (You select Greed for XXXX, etc)
	local selected = function(str, ty, own) handler(str, 
		function(who, item)
			if own then item = who who = UnitName('player') end
			--Dump({"---[Selected -> Handler -> Function]---", str, ty, item,who, "===[Selected -> Handler -> Function]===", ""}) 
			self:trigger('RollSelected', ty, who, item) 
		end) 
	end

	-- Rolled handler (XXXX roll - # by xxxxx, etc)
	local rolled = function(str, ty, own) handler(str, 
		function(roll, item, who) 
			self:trigger('RollMade', ty, ncheck(own, who), roll, item) 
		end)
	end

	-- Add handlers using construct functions, ordered specifically
	selected(LOOT_ROLL_GREED, 'greed')
	selected(LOOT_ROLL_NEED, 'need')
	selected(LOOT_ROLL_GREED_SELF, 'greed', true)
	selected(LOOT_ROLL_NEED_SELF, 'need', true)
	handler(LOOT_ROLL_ALL_PASSED, function(item) self:trigger('RollAllPassed', item) end)
	selected(LOOT_ROLL_PASSED, 'pass')
	selected(LOOT_ROLL_PASSED_SELF, 'pass', true)
	rolled(LOOT_ROLL_ROLLED_GREED, 'greed')
	rolled(LOOT_ROLL_ROLLED_NEED, 'need')
	handler(LOOT_ROLL_WON, function(who, item) self:trigger('RollWon', item, ncheck(own, who)) end)

	-- Possibly deactivate a older version (We don't have a deactivate! Nya!)
	if oldDeactivate then oldDeactivate(oldLib) end
end

-- Wrapper for triggering Ace events (We use a max of four args)
function lib:trigger(event, a, b, c, d)
	-- We debug it going out, let's do it again coming in!
	--Dump({ "---[SpecialEvents_"..event.."]---", a, b, c, d, "===[END EVENT]===", "" })
	self:TriggerEvent("SpecialEvents_"..event, a, b, c, d)
end

-- Chatmsg handler
function lib:Handler(text)
	-- Handle group loot mode only if we're in the mode and there's a roll going
	if true then --if GetLootMethod() == 'group' and self.rolls > 0 then
		-- Move through the patterns one by one, match against the message
		for k, v in ipairs(self.rollhandlers) do
			local pat, func = v[1], v[2]
			local m1, m2, m3, m4 = deformat(text, pat)
			-- Match was found, call the handler with all captured values
			if m1 then
				-- Debug to see what we have
				--Dump({"---[PATTERN MATCHED]---", "---["..text.."]---", "---["..pat.."]---", m1, m2, m3, m4, "===[END PATTERN MATCH]===", ""})
				return func(m1, m2, m3, m4)
			end
		end
	end
	
	-- Optimize by grouped/not and weight patterns by usage
	-- Check group mode (1 corresponds to ungrouped, 2 to grouped)
	local group = (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) and 2 or 1
	-- If we're not currently sorting for our current mode, or it's been too long since last sort, sort ourpatterns for efficiency
	if not self.currentsort or self.currentsort ~= group or self.unsortedloot > 25 then
		self.unsortedloot = 0
		self.currentsort = group
		table.sort(self.patterns, function(a, b) return a[group] < b[group] end)
	else
		self.unsortedloot = self.unsortedloot + 1
	end
	-- Iterate through patterns in our fancy sorted order.
	for i, v in ipairs(self.patterns) do
		-- Does it work? I don't know! XD
		local matches = self.matchcache
		matches = { deformat(text, v[3]) }
	-- [1] Self count [2] Group count [3] Pattern [4 key] Recipient [5 key] Item [6 key] Count (Our handy cheet-sheet)
		if matches[1] then
			-- We found a match, make note
			v[group] = v[group] + 1
			-- Scry recipient
			recipient = v[4] == "self" and UnitName("player") or matches[v[4]]
			-- And item....
			local item = v[5] == "coin" and "coin" or matches[v[5]]
			-- And how many...
			local count = item == "coin" and (self:ParseCoinString(matches[v[6]]) or 0) or (v[6] ~= nil and matches[v[6]] or 1)
			-- Tell the world!
			if item == "coin" then
				return self:TriggerEvent("SpecialEvents_CoinLooted", recipient, count, matches[v[6]])
			else
				return self:TriggerEvent("SpecialEvents_ItemLooted", recipient, count, item)
			end
		end
	end
end

-- Use our handy coinage reference from earlier, iterating through the coin types for a handy shortcut.
function lib:ParseCoinString(tstr)
	local total, tc = 0
	for k, v in pairs(self.coinage) do
		_, _, tc = tstr:find("(%d+) "..v[1])
		if tc then
			total = total + (tc * v[2])
		end
	end
	return total
end

--------------------------------
--      Load this bitch!      --
--------------------------------
AceLibrary:Register(lib, vmajor, vminor, activate)
